clear,close all;
%% import data
v = VideoReader('./data/906-17_l.avi'); %be sure to check if video location is correct before running program
video = zeros(v.Height,v.Width,3,round(v.Duration*v.FrameRate),'uint8'); %preallocate data for video, [height, width, color channels, frames]
i = 1;
while hasFrame(v) % extract all the frames in the video (for surveillance continuous data extraction needs to be applied)
    video(:,:,:,i) = readFrame(v);
%     imshow(video(:,:,:,i));drawnow;
    i = i+1;
end
% figure;colormap jet
%% Detect abnormal behaviour (Piotr's Master Toolbox, Bayesian Surprise)
Voptx = zeros(size(video,1),size(video,2),size(video,4)); %preallocate optical flow matrix [x-coordinate,y-coordinate,frame,optical flow direction (1 = x, 2 = y)]
Vopty = zeros(size(video,1),size(video,2),size(video,4));
[X,Y] = meshgrid(20:20:size(video,1)-20,20:20:size(video,2)-20); %initial particle positions (20 pixels between each particle, can be changed (be sure to change ROI and w_av as well))
[Xint,Yint] = meshgrid(20:20:size(video,1)-20,20:20:size(video,2)-20);
Opt_av = zeros(size(X,1),size(X,2),size(video,4),2);%preallocating
w_av = fspecial('gaussian',[21,21]); %filter for calculating the weighted average optical flow of a particle
Opt_loc= zeros(size(X,1),size(X,2),size(video,4),2);%preallocating
p=[0 0];%panic weight parameter: 0->individual characteristics, 1->crowd characteristics
tau=1;%relaxation parameter
V_des=zeros(size(Opt_loc));%preallocate
F_int=zeros(size(Opt_loc));%preallocate
S_f=zeros(size(X,1),size(X,2),size(video,4)); %preallocation
Opt_dif=zeros(size(X,1),size(X,2),size(video,4),2); %preallocating
SocialForce=zeros(size(X,1),size(X,2),size(video,4)); %preallocating
Vector.relsize=zeros(size(X,1),size(X,2),size(video,4)); %preallocating
Vector.angledif=zeros(size(X,1),size(X,2),size(video,4)); %preallocating
S=zeros(size(X,1),size(X,2),size(video,4)); %preallocating
w1=1; %weight factor of vector relative size
w2=1; %weight factor of vector angle difference
w3=1; %weight factor of Social Force
a=1; %tracking parameter
row2=[]; %initial empty matrix
col2=[]; %initial empty matrix
SurpriseThreshold=5;%surprise threshold after which program shows particle
%initialize
X2 = reshape(X,[1,size(X,1)*size(X,2)]); %reshape X into single array for plotting
Y2 = reshape(Y,[1,size(Y,1)*size(Y,2)]); %reshape Y into single array for plotting
I = video(:,:,:,1); %first video frame
figure; %open new figure window
imshow(I);hold on; %show first video frame
scatter(Y2,X2,5,'y','filled'); %show initial particle positions
hold off;
pause
%advection
tic %start stopwatch
for i = 2:size(video,4) %for every frame (start at frame 2 to initialize optical flow from frame 1 to 2.
    I1 = video(:,:,:,i-1); %previous frame
    I2 = video(:,:,:,i); %current frame
    [Vx,Vy] = opticalFlow(rgb2gray(I1),rgb2gray(I2),'smooth',1,'radius',10,'type','LK'); %use Piotr,s optical flow
    Voptx(:,:,i) = Vx; %reallocate optical flow x-vector into new frame dependent matrix
    Vopty(:,:,i) = Vy; %reallocate optical flow y-vector into new frame dependent matrix
    for x = 1:size(X,1) %particle x-coordinate
        for y = 1:size(X,2) %particle y-coordinate
            X_ROI = round([X(x,y)-10:X(x,y)+10]); %determine Region Of Interest
            Y_ROI = round([Y(x,y)-10:Y(x,y)+10]); %''
            if (min(X_ROI>0) && max(X_ROI) <= size(I,1)) && (min(Y_ROI >0) && max(Y_ROI) <= size(I,2)) %make sure the Region Of Interest is within the frame limits
                Opt_ROIx = Voptx(X_ROI,Y_ROI,i); % extract optical flow of ROI
                Opt_ROIy = Vopty(X_ROI,Y_ROI,i);
            else
                Opt_ROIx = zeros(size(w_av,1),size(w_av,2),1); %if Region of Interest is outside frame limits, set to zero array
                Opt_ROIy = zeros(size(w_av,1),size(w_av,2),1);
            end
            Opt_av(x,y,i,1) = sum(sum(w_av.*Opt_ROIx(:,:,1))); %calculate weighted average (x-direction)
            Opt_av(x,y,i,2) = sum(sum(w_av.*Opt_ROIy(:,:,1))); %calculate weighted average (y-direction)
            Opt_loc(x,y,i,1)= Voptx(round(X(x,y)),round(Y(x,y)),i); %Local particle velocity (x-direction)
            Opt_loc(x,y,i,2)= Vopty(round(X(x,y)),round(Y(x,y)),i); %Local particle velocity (y-direction)
            V_des(x,y,i,1)=(1-p(1)).*Opt_loc(x,y,i,1)+p(1).*Opt_av(x,y,i,1);%Desired x-velocity
            V_des(x,y,i,2)=(1-p(2)).*Opt_loc(x,y,i,2)+p(2).*Opt_av(x,y,i,2);%Desired y-velocity
            F_int(x,y,i,1)=(1./tau).*(V_des(x,y,i,1)-Opt_av(x,y,i,1))-(Opt_av(x,y,i,1)-Opt_av(x,y,i-1,1)); %estimate Interaction force x
            F_int(x,y,i,2)=(1./tau).*(V_des(x,y,i,2)-Opt_av(x,y,i,2))-(Opt_av(x,y,i,2)-Opt_av(x,y,i-1,2)); %estimate Interaction force y
            S_f(x,y,i)=S_f(x,y,i)+sqrt(F_int(x,y,i,1).^2+F_int(x,y,i,2).^2); %force flow
            Opt_dif(x,y,i,1)=Opt_av(x,y,i,1)-Opt_loc(x,y,i,1); %difference between average and local velocity (x-direction)
            Opt_dif(x,y,i,2)=Opt_av(x,y,i,2)-Opt_loc(x,y,i,2); %difference between avarage and local velocity (y-direction)
            Vector.relsize(x,y,i)=sqrt(Opt_dif(x,y,i,1).^2+Opt_dif(x,y,i,2).^2); %relative size Opt_dif to Opt_av
            Vector.angledif(x,y,i)=mod(abs(tan(Opt_av(x,y,i,2)./Opt_av(x,y,i,1))-tan(Opt_loc(x,y,i,2)./Opt_loc(x,y,i,1))),2*pi); %relative angular difference average direction to local direction (rad)
            SocialForce(x,y,i)=sqrt(F_int(x,y,i,1).^2+F_int(x,y,i,2).^2); %calculate Social Force
            S(x,y,i)=S(x,y,i-1)+w1.*Vector.relsize(x,y,i)+w2.*Vector.angledif(x,y,i)+w3.*SocialForce(x,y,i); %calculate weighed suspicion score
        end
    end
    if i==4*a %set up bayesian surprise, runsm function needed. Check every 4 frames
        a=a+1; %tracking parameter
        for x=1:size(X,1)
            for y=1:size(X,2) %scroll through x- and y-values
                row2=[]; %set empty row and column indexes for particle identification
                col2=[];
                Sps(x,y)=S(x,y,i)./i; %average surprise score increase per frame
                Spssave(x,y,a)=S(x,y,i)./i; %for analysing the score later on, for example for baseline analysing
                smod(x,y)=newsm(0.7); %set up Bayesian surprise algorithm
                smod(x,y).options.debug = 0; %turn of debugging option of Bayesian surprise
                smod(x,y).options.factordecay='yes'; %factor in a certain decay of relevant surprise values
                smod(x,y).options.alphaupdate='poisson'; %assume poisson distribution
                smod(x,y)=runsm(Sps(x,y),smod(x,y)); %run the actual Bayesian Surprise algorithm for each particle (x,y)
                surprise(x,y)=smod(x,y).surprise; %The surprise value taken from the Bayesian Surprise algorithm per particle
                surprisesave(x,y,a)=smod(x,y).surprise; %surprise value for analysing later on
                [row2,col2]=find(surprise>=SurpriseThreshold); %find the indexes for the particles over a certain threshold
                X(x,y) = X(x,y) + Opt_av(x,y,i,2); %new particle locations (X is actually the conventional y-direction and Y is actually the conventional x-direction)
                Y(x,y) = Y(x,y) + Opt_av(x,y,i,1);
            end
        end
    else
         for x=1:size(X,1)
             for y=1:size(X,2)
                 X(x,y) = X(x,y)+Opt_av(x,y,i,2);
                 Y(x,y) = Y(x,y)+Opt_av(x,y,i,1);
             end
         end
    end
    X2 = reshape(X,[1,size(X,1)*size(X,2)]); %reshape the X matrix into single X2 array, for drawing purposes
    Y2 = reshape(Y,[1,size(Y,1)*size(Y,2)]); %reshape the Y matrix into single Y2 array, for drawing purposes
    imshow(video(:,:,:,i)); hold on; %show current frame of the video
        if size(row2,1)>0 %checks if there is a particle with a surprise value above threshold
            for idx = 1:size(row2)
                scatter(Y(row2(idx),col2(idx)),X(row2(idx),col2(idx)),100,'y','o') %draw a circle around that particle
            end
        end
%     scatter(Y2,X2,5,'y','filled'); %uncomment to display particle location
	hold off;
    drawnow;
    [row,col]=find(round(X)>=(size(video,1)-10)); %reset particle location to initial position once they come too close to frame edge
       if size(row,1)>0
           for idx = 1:size(row) %particle x-coordinate
                X(row(idx),col(idx))=Xint(row(idx),col(idx));
                Y(row(idx),col(idx))=Yint(row(idx),col(idx));
           end
       end
       [row,col]=find(round(X)<=10);
       if size(row,1)>0
            for idx = 1:size(row) %particle x-coordinate
                X(row(idx),col(idx))=Xint(row(idx),col(idx));
                Y(row(idx),col(idx))=Yint(row(idx),col(idx));
            end
       end
       [row,col]=find(round(Y)>=(size(video,2)-10));
       if size(row,1)>0
            for idx = 1:size(row) %particle x-coordinate
                X(row(idx),col(idx))=Xint(row(idx),col(idx));
                Y(row(idx),col(idx))=Yint(row(idx),col(idx));
            end
       end
       [row,col]=find(round(Y)<=10);
       if size(row,1)>0
            for idx = 1:size(row) %particle x-coordinate
                X(row(idx),col(idx))=Xint(row(idx),col(idx));
                Y(row(idx),col(idx))=Yint(row(idx),col(idx));
            end
       end
end
toc %end stopwatch